import processing.serial.*;
import oscP5.*;
import netP5.*;

OscP5 oscP5;
NetAddress localhostAddress;
OscMessage oscMsg;
OscMessage daynightMsg;

byte[] inBuffer;
short distance = 80;
Serial port;

MothAnimation moth;
EnvironmentAnimation environment;

// processing's usual setup
void setup() {
  size(640, 480);
  // set up OSC
  oscP5 = new OscP5(this, 8001);
  localhostAddress = new NetAddress("127.0.0.1", 8888);
  oscMsg = new OscMessage("/distance");
  daynightMsg = new OscMessage("/daynight");
 
  moth = new MothAnimation();
  environment = new EnvironmentAnimation();
  
  println(Serial.list());
  port = new Serial(this, Serial.list()[2], 9600);
  
  inBuffer = new byte[4];
  
  frameRate(30);
  colorMode(RGB);
}


// processing's ever-repeating draw method
void draw() {
  background(0);
   
  environment.draw();
  moth.draw();
  
  if( port.available() > 0) {
    // ultrasonic sensor sends in 8 bytes of data (one word variable), but only the first byte counts. 
    // the rest seem to be 0
    port.readBytes(inBuffer);
    if (inBuffer != null) {
      if(true || distance != inBuffer[0]) {
        //if(inBuffer[0] > 0)
          println(inBuffer[0]+" "+inBuffer[1]+" "+inBuffer[2]+" "+inBuffer[3]);
        // println(inBuffer[8]+" "+inBuffer[9]+" "+inBuffer[10]+" "+inBuffer[11]+" "+inBuffer[12]+" "+inBuffer[13]+" "+inBuffer[14]+" "+inBuffer[15]+"  < ");
        
        if((inBuffer[1] > 20 || inBuffer[1] < 0) && environment.isDay) {
          print("start darkness!");
          environment.darknessFalls();
        } else if(inBuffer[1] < 15 && !environment.isDay) {
          print("let there be light!");
          environment.sunRises();
        }
        
        oscMsg.clear();
        oscMsg.setAddrPattern("/distance");
        oscMsg.add((int)inBuffer[0]); /* add an int to the osc message */
        oscP5.send(oscMsg, localhostAddress);
      } else
        ;
      
      distance = inBuffer[0];
    }
  }
}


/* shamelessly ripped from http://www.captain.at/howto-java-convert-binary-data.php */
public static double arr2double (byte[] arr, int start) {
  int i = 0;
  int len = 8;
  int cnt = 0;
  byte[] tmp = new byte[len];
  for (i = start; i < (start + len); i++) {
    tmp[cnt] = arr[i];
    //System.out.println(java.lang.Byte.toString(arr[i]) + " " + i);
    cnt++;
  }
  long accum = 0;
  i = 0;
  for ( int shiftBy = 0; shiftBy < 64; shiftBy += 8 ) {
    accum |= ( (long)( tmp[i] & 0xff ) ) << shiftBy;
    i++;
  }
  return Double.longBitsToDouble(accum);
}




/* a class to manage the moth animation */
class MothAnimation {
  PImage[] moths;
  short curMothFrame = 0;
  int mothLastFrameTime = 0;
  float x, y, w, h;
  float flightAngle;
  float flightSpeed;
  boolean isDayTime = true;
  
  public MothAnimation() {
    println("creating the moth");
    moths = new PImage [6];
    moths[0] = loadImage("mot1.png");
    moths[1] = loadImage("mot2.png");
    moths[2] = loadImage("mot3.png");
    moths[3] = loadImage("mot4.png");
    moths[4] = loadImage("mot3.png");
    moths[5] = loadImage("mot2.png");
    w = 102; h = 77; flightSpeed = 3; flightAngle = 0;
    x = width/2; y = height/2;
  }
  
  public void draw() {
    
    if(curMothFrame == 5) curMothFrame = 0;
    else curMothFrame++;
      
    // fly in a "cirlish" motion
    flightSpeed = random(30) + 5;
     
    w = (20480 / (distance * distance + 1)    + random(5)  ); 
    h = (15360  / (distance * distance + 1)    + random(5)  );
      
    x += flightSpeed * cos(flightAngle);
    y += flightSpeed * sin(flightAngle);
      
    if(x > width) {
      flightAngle = PI;
      flightSpeed = 20;
    } else if(x < 10) {
      flightAngle = 0;
      flightSpeed = 20;
    }
      
    if(y > height) {
      flightAngle = 1.5 * PI;
      flightSpeed = 20;
    } else if(y < 10) {
      flightAngle = .5 * PI;
      flightSpeed = 20;
    } else if(frameCount % 5 == 0) {
      flightAngle +=  random(1.0) - .5;
    }
    
    if(frameCount % 5 != 0)   
      //blend( moths[curMothFrame], 0, 0, 1024, 768, (int)(x - w/2), (int)(y - h/2), (int)w, (int)h, BLEND); 
      image( moths[curMothFrame], x - w/2, y - h/2, w, h );
  } 
}


class EnvironmentAnimation {
  PImage dayImg;
  PImage nightImg;
  PImage clouds;
  int cloudPos = -3900;
  PImage nightClouds;
  PImage[] noiseFrames;
  int curNoiseFrame;
  
  int imgh, imgw;
  boolean isDay = false;
  boolean transitionDone = true;
  int val = 0;
  
  public EnvironmentAnimation() {
    // 2592 * 1944
    dayImg = loadImage("Day.png");
    nightImg = loadImage("Night.png");
    clouds = loadImage("clouds.png");
    nightClouds = loadImage("Night_clouds.png");
    imgw = 2592;
    imgh = 1944;
    noiseFrames = new PImage[3];
    noiseFrames[0] = loadImage("2noise1.jpg");
    makeTransparent(noiseFrames[0]);
    noiseFrames[1] = loadImage("2noise2.jpg");
    makeTransparent(noiseFrames[1]);
    noiseFrames[2] = loadImage("2noise3.jpg");
    makeTransparent(noiseFrames[2]);
  }
  
  private void makeTransparent(PImage i) {
    i.loadPixels();
    for(int p = 0; p < i.width * i.height; p++) {
      i.pixels[p] = color( red(i.pixels[p]), green(i.pixels[p]), blue(i.pixels[p]), 64); 
    }
    i.updatePixels();
  }
  
  public void draw() {
    noStroke();
    
    // background(val);
    
    if(isDay) {
      //fill(255-val, 20, val);
      //rect(0, 0, 640, 400);
          
      // we're progressing towards day
      if(!transitionDone) {  
        if(val < 254) {
          //tint(255, 255 - val);
          image(nightClouds, cloudPos, 0, 4540, 780);
          image(nightImg, 0, 0, width, height);
          
          tint(255, val);
          //blend(dayImg, 0, 0, dayImg.width, dayImg.height, 0, 0, width, height, BLEND);
          image(clouds, cloudPos, 0, 4540, 780);
          image(dayImg, 0, 0, width, height);
          noTint();
          
          val += 5;
        } else {
          transitionDone = true;
          image(clouds, cloudPos, 0, 4540, 780);
          image(dayImg, 0, 0, width, height);
        }
      }
      else {
        image(clouds, cloudPos, 0, 4540, 780);
        image(dayImg, 0, 0, width, height);
      }
    } 
    else {
      //fill(255 - val, 20, val);
      //rect(0, 0, 640, 400);
      
      // it's getting dark..
      if(!transitionDone) {  
        if(val > 1) {
          //tint(255, val);
          image(clouds, cloudPos, 0, 4540, 780);
          image(dayImg, 0, 0, width, height);
          
          tint(255, 255 - val);
          //blend(nightImg, 0, 0, nightImg.width, nightImg.height, 0, 0, width, height, BLEND);
          image(nightClouds, cloudPos, 0, 4540, 780);
          image(nightImg, 0, 0, width, height);
          noTint();
          val -= 5;
        } else {
          transitionDone = true;
          image(nightClouds, cloudPos, 0, 4540, 780);
          image(nightImg, 0, 0, width, height);
        } 
      } 
      else {
        image(nightClouds, cloudPos, 0, 4540, 780);
        image(nightImg, 0, 0, width, height);
      }
    }
    
    // move da cloud
    if(frameCount % 5 == 0) {
      if(cloudPos == -3900)
        cloudPos = 0;
      else
        cloudPos -= (1 + 10/distance);
    }
    
    // do da noise!
    int r = (int)random(100) - 50;
    blend(noiseFrames[round(curNoiseFrame)], 0, 0, width, height, 0, 0, width, height, OVERLAY);
    if(frameCount % 3 == 0)
      curNoiseFrame = round( random(2.4) );
  }
  
  public void sunRises() {
    if(transitionDone) { 
      isDay = true; 
      transitionDone = false;
      daynightMsg.clear();
      daynightMsg.setAddrPattern("/daynight");
      daynightMsg.add(1); /* add an int to the osc message */
      oscP5.send(daynightMsg, localhostAddress); 
    }
  }
  
  public void darknessFalls() {
    if(transitionDone) {
      isDay = false;
      transitionDone = false;
      daynightMsg.clear();
      daynightMsg.setAddrPattern("/daynight");
      daynightMsg.add(0); /* add an int to the osc message */
      oscP5.send(daynightMsg, localhostAddress);
    }
  }
}


void keyPressed() {
  println("key pressed");
  if(key == 'w') {
    moth.w *= 1.3; 
    moth.h *= 1.3;
  } else if(key == 's') {
    moth.w *= 0.8;
    moth.h *= 0.8;
  } else if(key == 'a') {
    environment.sunRises();
  } else if(key == 'd') {
    environment.darknessFalls();
  }
}
